==> agent_memory_jojo-0.1.1-py3-none-any.whl: agent_memory_jojo/processors/intelligent.py <==
import uuid
from langchain_core.messages import HumanMessage, AIMessage
from ..client.qdrant_client import QdrantVectorDB
from ..agents.agent import Agent
from ..schema.reflection_schema import ReflectionSchema
from ..schema.document_schema import DocumentPayload
from ..schema.search_schema import MatchAnyOrInterval
from ..utils.app_utils import AppUtil
from pathlib import Path
from pydantic import BaseModel
from typing import Dict, Optional, Union, Any, List
import json
from ..config.config_helper import Configuration
from ..config.logger import Logger

logger = Logger(__name__)

settings = Configuration().get_config('qdrant')

class MemoryFormatter:
    @staticmethod
    def create_payload(
        id: uuid.UUID,
        user_id: str,
        collection_name: str,
        data: Dict[str, Any]
    ) -> DocumentPayload:
        if not id:
            id=str(uuid.uuid4())
        
        return DocumentPayload(
            id=id,
            user_id=user_id,
            summary=data['summary'],
            document_type="preference",
            memory_type="long_term_memory",
            collection_name=collection_name
        )


class IntelligentMemoryManager:
    BASE_DIR = Path(__file__).resolve().parent.parent
    PROMPT_DIR = BASE_DIR / "agents" / "prompt" / "templates"
    keys_to_ignore = ['user_id', 'document_type', 'memory_type', 'collection_name']

    def __init__(self, user_id: str, collection_name: str = settings['collection_name']) -> None:
        self.collection_name = collection_name
        self.vectordb = QdrantVectorDB(collection_name=self.collection_name)
        self.limit = self.vectordb.limit
        self.returned_memories = []
        self.user_id = user_id
        self.initialize_system_messages()
        self.initialize_agents()
    
    def initialize_agents(self) -> None:
        """Initializes the reflection agent with the specified schema and system prompt."""
        self.reflection_agent = self.create_agent(
            Agent,
            name="AI agent",
            system_message=self.reflection_system_prompt,
            schema=ReflectionSchema
        )
        self.fact_retrieval_agent=Agent(
            name="smart memory manager",
            schema=None,
            system_message=self.fact_retrieval_system_prompt
        )
        self.reflection_instance = self.reflection_agent.get_agent()
    
    def initialize_system_messages(self) -> None:
        """Loads system prompts from specified files."""
        self.reflection_system_prompt = AppUtil.load_file(
            self.PROMPT_DIR / 'reflection_system_prompt.txt'
        )
        self.fact_retrieval_system_prompt = AppUtil.load_file(
            self.PROMPT_DIR / 'fact_retrieval_system_prompt.txt'
        )

    def create_agent(
        self,
        agent_class,
        name: str,
        system_message: str,
        schema: Optional[BaseModel] = None,
        **kwargs
    ) -> Agent:
        """
        Creates an instance of the specified agent.

        Args:
            agent_class: The agent class to instantiate.
            name (str): The name of the agent.
            system_message (str): The system message指导 agent behavior.
            schema (Optional[BaseModel], optional): The schema for structuring agent outputs.
                Defaults to None.

        Returns:
            Agent: An instance of the specified agent.
        """
        return agent_class(
            name=name,
            system_message=system_message,
            schema=schema,
            **kwargs
        )
    
    def get_document(self, data: Dict[str, Any], id=None) -> List[Dict[str, Any]]:
        """
        Creates and returns a document based on the input data.

        Args:
            data (Dict[str, Any]): Data to be used in document creation.

        Returns:
            List[Dict[str, Any]]: A list containing the formatted document.
        """
        payload = MemoryFormatter.create_payload(
            id=id,
            user_id=self.user_id,
            collection_name=self.collection_name,
            data=data
        )
        formatted_content = payload.summary
        return [{"page_content": formatted_content,"metadata": payload.dict(exclude_none=True)}]
    
    def add(self, memory_data: Dict[str, Any]) -> None:
        """
        Adds a new memory to the vector database after processing through the reflection agent.

        Args:
            memory_data (Dict[str, Any]): The memory data to be stored.
        """
        try:
            # Generate facts based on the conversation between the user and the agent
            reflection = self.reflection_instance.invoke({
                "messages": [HumanMessage(content=json.dumps(memory_data))]
            }).model_dump()
            # current_fact = {
            #     **reflection,
            #     'user_search': 'N/A'
            # }
            current_fact = reflection
            
            # Get the related memory based on the newly generated facts
            new_generated_search_facts = [document['page_content'] for document in self.get_document(data=current_fact)][0]

            existing_memories = self.query(
                query=new_generated_search_facts
            )
            retrieved_old_memory = [{
                k: v for k, v in metadata.items() if k not in self.keys_to_ignore
            } for metadata in [
                document.metadata for document in existing_memories
            ]]
            logger.info(f"Total existing memories: {len(retrieved_old_memory)}")

            # mapping UUIDs with integers for handling UUID hallucinations
            temp_uuid_mapping = {}
            for idx, item in enumerate(retrieved_old_memory):
                temp_uuid_mapping[str(idx)] = item["id"]
                retrieved_old_memory[idx]["id"] = str(idx)

            self.fact_retrieval_agent.set(retrieved_old_memory=retrieved_old_memory)
            self.fact_retrieval_agent.set(response_content=current_fact)
            self.fact_retrieval_active = self.fact_retrieval_agent.get_agent()

            new_memories_with_actions = self.fact_retrieval_active.invoke(
                {
                    "messages":[
                        HumanMessage(
                            content="what is the New Memory Element with their respective event, the event is a must for each element, this output must be in JSON format"
                        )
                    ]
                }
            ).content
            new_memories_with_actions = AppUtil.remove_code_blocks(new_memories_with_actions)
            try:
                new_memories_with_actions = json.loads(new_memories_with_actions)
            except json.JSONDecodeError as ex:
                logger.error(f"Error decoding JSON: {ex}")
            
            try:
                for result in new_memories_with_actions["memory"]:
                    logger.info(result)
                    output = result
                    try:
                        if output["event"] == "ADD":
                            output.pop('id', None)
                            event = output.pop('event', None)
                            output.pop('event', None)
                            memory_id = self._create_memory(
                                data=output
                            )
                            self.returned_memories.append(
                                {
                                    "id": memory_id,
                                    "memory": output,
                                    "event": event,
                                }
                            )
                        elif output["event"] == "UPDATE":
                            event = output.pop('event', None)
                            output.pop('event', None)
                            old_memory = self._update_memory(
                                memory_id=temp_uuid_mapping[output["id"]],
                                data=output
                            )
                            self.returned_memories.append(
                                {
                                    "id": temp_uuid_mapping[output["id"]],
                                    "memory": output,
                                    "event": event,
                                    "previous_memory": old_memory
                                }
                            )
                        elif output["event"] == "DELETE":
                            event = output.pop('event', None)
                            output.pop('event', None)
                            self._delete_memory(memory_id=temp_uuid_mapping[output["id"]])
                            self.returned_memories.append(
                                {
                                    "id": temp_uuid_mapping[output["id"]],
                                    "memory": output,
                                    "event": event,
                                }
                            )
                        elif output["event"] == "NONE":
                            logger.info("NOOP for Memory.")
                    except Exception as e:
                        logger.error(f"Error in new_memories_with_actions: {e}")
            except Exception as e:
                logger.error(f"Error in new_memories_with_actions: {e}")
            
        except Exception as e:
            logger.error(f"Error adding memory: {str(e)}")
            raise

    def _update_memory(self, memory_id, data):
        logger.info(f"Updating memory with {data=}")
        try:
            existing_memory = self.get(memory_id)
        except Exception:
            raise ValueError(f"Error getting memory with ID {memory_id}. Please provide a valid 'memory_id'")

        new_metadata = {}
        new_metadata["data"] = data
        for exist_mem in existing_memory:
            if "id" in exist_mem:
                #Prevent resetting the id
                new_metadata['data']['id'] = exist_mem['id']
            if "user_id" in exist_mem:
                #Prevent resetting the user id
                new_metadata['data']['user_id'] = exist_mem['user_id']
        
        id = new_metadata['data'].pop('id', None)
        new_metadata['data'].pop('id', None)
        to_embed = self.get_document(data=new_metadata['data'], id=id)
        self.vectordb.inject(to_embed)

        old_memory = [{
            k: v for k, v in obj.items() if k not in self.keys_to_ignore
        } for obj in [
            document for document in existing_memory
        ]]

        return old_memory
    
    def get_relevant_operation(self):
        return {"result": self.returned_memories}

    def _create_memory(self, data):
        data_document = self.get_document(data=data)
        self.vectordb.inject(data_document, is_batch=False)
        memory_id = [document['metadata'] for document in data_document][0]['id']
        
        return memory_id
    
    def _delete_memory(self, memory_id):
        self.vectordb.delete_one([memory_id])
        return {"message": f"Memory with id {memory_id} deleted successfully!"}
    
    def get(self, id):
        records = self.vectordb.get(id)
        keys_to_ignore = [key for key in keys_to_ignore if key != 'user_id']
        # Remove specified keys from each metadata dictionary
        metadata = [
            {k: v for k, v in metadata['metadata'].items() if k not in keys_to_ignore}
            for metadata in [record.payload for record in records]
        ]
        return metadata
    
    def query(self, query: str) -> List[Dict[str, Any]]:
        """
        Retrieves memories from the vector database based on the provided query.

        Args:
            query (str): The search query to filter memories.

        Returns:
            List[Dict[str, Any]]: List of retrieved memories.
        """
        filters = {
            'user_id': MatchAnyOrInterval(any=[str(self.user_id)])
        }
        return self.vectordb.search(query, limit=self.limit, filters=filters)
    
    def search(self, query: str) -> List[Dict[str, Any]]:
        output = self.query(query)
        metadata_list = [{
            k: v for k, v in metadata.items() if k not in self.keys_to_ignore
        } for metadata in [document.metadata for document in output]]

        return metadata_list

==> agent_memory_jojo-0.1.2-py3-none-any.whl: agent_memory_jojo/processors/intelligent.py <==
import os
import dotenv
dotenv.load_dotenv(dotenv.find_dotenv(), override=True)
import uuid
from langchain_core.messages import HumanMessage, AIMessage
from ..client.qdrant_client import QdrantVectorDB
from ..agents.agent import Agent
from ..schema.reflection_schema import ReflectionSchema
from ..schema.document_schema import DocumentPayload
from ..schema.search_schema import MatchAnyOrInterval
from ..utils.app_utils import AppUtil
from pathlib import Path
from pydantic import BaseModel
from typing import Dict, Optional, Union, Any, List
import json
from ..config.config_helper import Configuration
from ..config.logger import Logger

logger = Logger(__name__)

settings = Configuration().get_config('qdrant')

class MemoryFormatter:
    @staticmethod
    def create_payload(
        id: uuid.UUID,
        user_id: str,
        collection_name: str,
        data: Dict[str, Any]
    ) -> DocumentPayload:
        if not id:
            id=str(uuid.uuid4())
        
        return DocumentPayload(
            id=id,
            user_id=user_id,
            summary=data['summary'],
            document_type="preference",
            memory_type="long_term_memory",
            collection_name=collection_name
        )


class IntelligentMemoryManager:
    BASE_DIR = Path(__file__).resolve().parent.parent
    PROMPT_DIR = BASE_DIR / "agents" / "prompt" / "templates"
    keys_to_ignore = ['user_id', 'document_type', 'memory_type', 'collection_name']

    def __init__(self, user_id: str, collection_name: str = os.environ.get("COLLECTION_NAME")) -> None:
        self.collection_name = collection_name
        self.vectordb = QdrantVectorDB(collection_name=self.collection_name)
        self.limit = self.vectordb.limit
        self.returned_memories = []
        self.user_id = user_id
        self.initialize_system_messages()
        self.initialize_agents()
    
    def initialize_agents(self) -> None:
        """Initializes the reflection agent with the specified schema and system prompt."""
        self.reflection_agent = self.create_agent(
            Agent,
            name="AI agent",
            system_message=self.reflection_system_prompt,
            schema=ReflectionSchema
        )
        self.fact_retrieval_agent=Agent(
            name="smart memory manager",
            schema=None,
            system_message=self.fact_retrieval_system_prompt
        )
        self.reflection_instance = self.reflection_agent.get_agent()
    
    def initialize_system_messages(self) -> None:
        """Loads system prompts from specified files."""
        self.reflection_system_prompt = AppUtil.load_file(
            self.PROMPT_DIR / 'reflection_system_prompt.txt'
        )
        self.fact_retrieval_system_prompt = AppUtil.load_file(
            self.PROMPT_DIR / 'fact_retrieval_system_prompt.txt'
        )

    def create_agent(
        self,
        agent_class,
        name: str,
        system_message: str,
        schema: Optional[BaseModel] = None,
        **kwargs
    ) -> Agent:
        """
        Creates an instance of the specified agent.

        Args:
            agent_class: The agent class to instantiate.
            name (str): The name of the agent.
            system_message (str): The system message指导 agent behavior.
            schema (Optional[BaseModel], optional): The schema for structuring agent outputs.
                Defaults to None.

        Returns:
            Agent: An instance of the specified agent.
        """
        return agent_class(
            name=name,
            system_message=system_message,
            schema=schema,
            **kwargs
        )
    
    def get_document(self, data: Dict[str, Any], id=None) -> List[Dict[str, Any]]:
        """
        Creates and returns a document based on the input data.

        Args:
            data (Dict[str, Any]): Data to be used in document creation.

        Returns:
            List[Dict[str, Any]]: A list containing the formatted document.
        """
        payload = MemoryFormatter.create_payload(
            id=id,
            user_id=self.user_id,
            collection_name=self.collection_name,
            data=data
        )
        formatted_content = payload.summary
        return [{"page_content": formatted_content,"metadata": payload.dict(exclude_none=True)}]
    
    def add(self, memory_data: Dict[str, Any]) -> None:
        """
        Adds a new memory to the vector database after processing through the reflection agent.

        Args:
            memory_data (Dict[str, Any]): The memory data to be stored.
        """
        try:
            # Generate facts based on the conversation between the user and the agent
            reflection = self.reflection_instance.invoke({
                "messages": [HumanMessage(content=json.dumps(memory_data))]
            }).model_dump()
            # current_fact = {
            #     **reflection,
            #     'user_search': 'N/A'
            # }
            current_fact = reflection
            
            # Get the related memory based on the newly generated facts
            new_generated_search_facts = [document['page_content'] for document in self.get_document(data=current_fact)][0]

            existing_memories = self.query(
                query=new_generated_search_facts
            )
            retrieved_old_memory = [{
                k: v for k, v in metadata.items() if k not in self.keys_to_ignore
            } for metadata in [
                document.metadata for document in existing_memories
            ]]
            logger.info(f"Total existing memories: {len(retrieved_old_memory)}")

            # mapping UUIDs with integers for handling UUID hallucinations
            temp_uuid_mapping = {}
            for idx, item in enumerate(retrieved_old_memory):
                temp_uuid_mapping[str(idx)] = item["id"]
                retrieved_old_memory[idx]["id"] = str(idx)

            self.fact_retrieval_agent.set(retrieved_old_memory=retrieved_old_memory)
            self.fact_retrieval_agent.set(response_content=current_fact)
            self.fact_retrieval_active = self.fact_retrieval_agent.get_agent()

            new_memories_with_actions = self.fact_retrieval_active.invoke(
                {
                    "messages":[
                        HumanMessage(
                            content="what is the New Memory Element with their respective event, the event is a must for each element, this output must be in JSON format"
                        )
                    ]
                }
            ).content
            new_memories_with_actions = AppUtil.remove_code_blocks(new_memories_with_actions)
            try:
                new_memories_with_actions = json.loads(new_memories_with_actions)
            except json.JSONDecodeError as ex:
                logger.error(f"Error decoding JSON: {ex}")
            
            try:
                for result in new_memories_with_actions["memory"]:
                    logger.info(result)
                    output = result
                    try:
                        if output["event"] == "ADD":
                            output.pop('id', None)
                            event = output.pop('event', None)
                            output.pop('event', None)
                            memory_id = self._create_memory(
                                data=output
                            )
                            self.returned_memories.append(
                                {
                                    "id": memory_id,
                                    "memory": output,
                                    "event": event,
                                }
                            )
                        elif output["event"] == "UPDATE":
                            event = output.pop('event', None)
                            output.pop('event', None)
                            old_memory = self._update_memory(
                                memory_id=temp_uuid_mapping[output["id"]],
                                data=output
                            )
                            self.returned_memories.append(
                                {
                                    "id": temp_uuid_mapping[output["id"]],
                                    "memory": output,
                                    "event": event,
                                    "previous_memory": old_memory
                                }
                            )
                        elif output["event"] == "DELETE":
                            event = output.pop('event', None)
                            output.pop('event', None)
                            self._delete_memory(memory_id=temp_uuid_mapping[output["id"]])
                            self.returned_memories.append(
                                {
                                    "id": temp_uuid_mapping[output["id"]],
                                    "memory": output,
                                    "event": event,
                                }
                            )
                        elif output["event"] == "NONE":
                            logger.info("NOOP for Memory.")
                    except Exception as e:
                        logger.error(f"Error in new_memories_with_actions: {e}")
            except Exception as e:
                logger.error(f"Error in new_memories_with_actions: {e}")
            
        except Exception as e:
            logger.error(f"Error adding memory: {str(e)}")
            raise

    def _update_memory(self, memory_id, data):
        logger.info(f"Updating memory with {data=}")
        try:
            existing_memory = self.get(memory_id)
        except Exception:
            raise ValueError(f"Error getting memory with ID {memory_id}. Please provide a valid 'memory_id'")

        new_metadata = {}
        new_metadata["data"] = data
        for exist_mem in existing_memory:
            if "id" in exist_mem:
                #Prevent resetting the id
                new_metadata['data']['id'] = exist_mem['id']
            if "user_id" in exist_mem:
                #Prevent resetting the user id
                new_metadata['data']['user_id'] = exist_mem['user_id']
        
        id = new_metadata['data'].pop('id', None)
        new_metadata['data'].pop('id', None)
        to_embed = self.get_document(data=new_metadata['data'], id=id)
        self.vectordb.inject(to_embed)

        old_memory = [{
            k: v for k, v in obj.items() if k not in self.keys_to_ignore
        } for obj in [
            document for document in existing_memory
        ]]

        return old_memory
    
    def get_relevant_operation(self):
        return {"result": self.returned_memories}

    def _create_memory(self, data):
        data_document = self.get_document(data=data)
        self.vectordb.inject(data_document, is_batch=False)
        memory_id = [document['metadata'] for document in data_document][0]['id']
        
        return memory_id
    
    def _delete_memory(self, memory_id):
        self.vectordb.delete_one([memory_id])
        return {"message": f"Memory with id {memory_id} deleted successfully!"}
    
    def get(self, id):
        records = self.vectordb.get(id)
        keys_to_ignore = [key for key in keys_to_ignore if key != 'user_id']
        # Remove specified keys from each metadata dictionary
        metadata = [
            {k: v for k, v in metadata['metadata'].items() if k not in keys_to_ignore}
            for metadata in [record.payload for record in records]
        ]
        return metadata
    
    def query(self, query: str) -> List[Dict[str, Any]]:
        """
        Retrieves memories from the vector database based on the provided query.

        Args:
            query (str): The search query to filter memories.

        Returns:
            List[Dict[str, Any]]: List of retrieved memories.
        """
        filters = {
            'user_id': MatchAnyOrInterval(any=[str(self.user_id)])
        }
        return self.vectordb.search(query, limit=self.limit, filters=filters)
    
    def search(self, query: str) -> List[Dict[str, Any]]:
        output = self.query(query)
        metadata_list = [{
            k: v for k, v in metadata.items() if k not in self.keys_to_ignore
        } for metadata in [document.metadata for document in output]]

        return metadata_list
